@@ -1,31 +1,13 @@
 package org.codehaus.mojo.exec;
 
-/*
- * Licensed to the Apache Software Foundation (ASF) under one
- * or more contributor license agreements.  See the NOTICE file
- * distributed with this work for additional information
- * regarding copyright ownership.  The ASF licenses this file
- * to you under the Apache License, Version 2.0 (the
- * "License"); you may not use this file except in compliance
- * with the License.  You may obtain a copy of the License at
- *
- *     http://www.apache.org/licenses/LICENSE-2.0
- *
- * Unless required by applicable law or agreed to in writing,
- * software distributed under the License is distributed on an
- * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
- * KIND, either express or implied.  See the License for the
- * specific language governing permissions and limitations
- * under the License.
- */
-
-import java.io.File;
+import java.io.IOException;
+import java.lang.invoke.MethodHandle;
+import java.lang.invoke.MethodHandles;
+import java.lang.invoke.MethodType;
 import java.lang.reflect.InvocationTargetException;
-import java.lang.reflect.Method;
-import java.lang.reflect.Modifier;
-import java.net.MalformedURLException;
-import java.net.URL;
 import java.net.URLClassLoader;
+import java.nio.file.Path;
+import java.nio.file.Paths;
 import java.util.ArrayList;
 import java.util.Collection;
 import java.util.Collections;
@@ -33,23 +15,23 @@ import java.util.HashSet;
 import java.util.List;
 import java.util.Properties;
 import java.util.Set;
+import java.util.concurrent.CountDownLatch;
+import java.util.concurrent.ExecutorService;
+import java.util.concurrent.ForkJoinPool;
 
 import org.apache.maven.artifact.Artifact;
-import org.apache.maven.artifact.factory.ArtifactFactory;
-import org.apache.maven.artifact.metadata.ArtifactMetadataSource;
-import org.apache.maven.artifact.repository.ArtifactRepository;
+import org.apache.maven.artifact.resolver.ArtifactResolutionException;
+import org.apache.maven.artifact.resolver.ArtifactResolutionRequest;
 import org.apache.maven.artifact.resolver.ArtifactResolutionResult;
-import org.apache.maven.artifact.resolver.ArtifactResolver;
-import org.apache.maven.model.Dependency;
+import org.apache.maven.artifact.resolver.ResolutionErrorHandler;
 import org.apache.maven.plugin.MojoExecutionException;
 import org.apache.maven.plugin.MojoFailureException;
 import org.apache.maven.plugins.annotations.Component;
 import org.apache.maven.plugins.annotations.Mojo;
 import org.apache.maven.plugins.annotations.Parameter;
 import org.apache.maven.plugins.annotations.ResolutionScope;
-import org.apache.maven.project.MavenProject;
-import org.apache.maven.project.MavenProjectBuilder;
-import org.apache.maven.project.artifact.MavenMetadataSource;
+import org.apache.maven.project.ProjectBuilder;
+import org.apache.maven.repository.RepositorySystem;
 
 /**
  * Executes the supplied java class in the current VM with the enclosing project's dependencies as classpath.
@@ -62,46 +44,40 @@ public class ExecJavaMojo
     extends AbstractExecMojo
 {
     @Component
-    private ArtifactResolver artifactResolver;
+    private RepositorySystem repositorySystem;
 
     @Component
-    private ArtifactFactory artifactFactory;
+    private ResolutionErrorHandler resolutionErrorHandler;
 
-    @Component
-    private ArtifactMetadataSource metadataSource;
-
-    /**
-     * @since 1.0
-     */
-    @Parameter( readonly = true, required = true, defaultValue = "${localRepository}" )
-    private ArtifactRepository localRepository;
-
-    /**
-     * @since 1.1-beta-1
-     */
-    @Parameter( readonly = true, required = true, defaultValue = "${project.remoteArtifactRepositories}" )
-    private List<ArtifactRepository> remoteRepositories;
 
     /**
      * @since 1.0
      */
     @Component
-    private MavenProjectBuilder projectBuilder;
+    private ProjectBuilder projectBuilder;
 
     /**
-     * @since 1.1-beta-1
-     */
-    @Parameter( readonly = true, defaultValue = "${plugin.artifacts}" )
-    private List<Artifact> pluginDependencies;
-
-    /**
-     * The main class to execute.
+     * The main class to execute.<br>
+     * With Java 9 and above you can prefix it with the modulename, e.g. <code>com.greetings/com.greetings.Main</code>
+     * Without modulename the classpath will be used, with modulename a new modulelayer will be created.
      * 
      * @since 1.0
      */
     @Parameter( required = true, property = "exec.mainClass" )
     private String mainClass;
 
+
+    /**
+     * Forces the creation of fork join common pool to avoids the threads to be owned by the isolated thread group
+     * and prevent a proper shutdown.
+     * If set to zero the default parallelism is used to precreate all threads,
+     * if negative it is ignored else the value is the one used to create the fork join threads.
+     *
+     * @since 3.0.1
+     */
+    @Parameter( property = "exec.preloadCommonPool", defaultValue = "0" )
+    private int preloadCommonPool;
+
     /**
      * The class arguments.
      * 
@@ -218,6 +194,15 @@ public class ExecJavaMojo
     @Parameter
     private List<String> additionalClasspathElements;
 
+    /**
+     * List of file to exclude from the classpath.
+     * It matches the jar name, for example {@code slf4j-simple-1.7.30.jar}.
+     *
+     * @since 3.0.1
+     */
+    @Parameter
+    private List<String> classpathFilenameExclusions;
+
     /**
      * Execute goal.
      * 
@@ -259,29 +244,41 @@ public class ExecJavaMojo
             getLog().debug( msg );
         }
 
+        if ( preloadCommonPool >= 0 )
+        {
+            preloadCommonPool();
+        }
+
         IsolatedThreadGroup threadGroup = new IsolatedThreadGroup( mainClass /* name */ );
         Thread bootstrapThread = new Thread( threadGroup, new Runnable()
         {
             public void run()
             {
+                int sepIndex = mainClass.indexOf( '/' );
+
+                final String bootClassName;
+                if ( sepIndex >= 0 )
+                {
+                    bootClassName = mainClass.substring( sepIndex + 1 );
+                }
+                else 
+                {
+                    bootClassName = mainClass;
+                }
+                
                 try
                 {
-                    Method main =
-                        Thread.currentThread().getContextClassLoader().loadClass( mainClass ).getMethod( "main",
-                                                                                                         new Class[] {
-                                                                                                             String[].class } );
-                    if ( !main.isAccessible() )
-                    {
-                        getLog().debug( "Setting accessibility to true in order to invoke main()." );
-                        main.setAccessible( true );
-                    }
-                    if ( !Modifier.isStatic( main.getModifiers() ) )
-                    {
-                        throw new MojoExecutionException( "Can't call main(String[])-method because it is not static." );
-                    }
-                    main.invoke( null, new Object[] { arguments } );
+                    Class<?> bootClass = Thread.currentThread().getContextClassLoader().loadClass( bootClassName );
+                    
+                    MethodHandles.Lookup lookup = MethodHandles.lookup();
+
+                    MethodHandle mainHandle =
+                        lookup.findStatic( bootClass, "main",
+                                                 MethodType.methodType( void.class, String[].class ) );
+                    
+                    mainHandle.invoke( arguments );
                 }
-                catch ( NoSuchMethodException e )
+                catch ( IllegalAccessException | NoSuchMethodException | NoSuchMethodError e )
                 { // just pass it on
                     Thread.currentThread().getThreadGroup().uncaughtException( Thread.currentThread(),
                                                                                new Exception( "The specified mainClass doesn't contain a main method with appropriate signature.",
@@ -292,13 +289,14 @@ public class ExecJavaMojo
                    Throwable exceptionToReport = e.getCause() != null ? e.getCause() : e;
                    Thread.currentThread().getThreadGroup().uncaughtException( Thread.currentThread(), exceptionToReport );
                 }
-                catch ( Exception e )
+                catch ( Throwable e )
                 { // just pass it on
                     Thread.currentThread().getThreadGroup().uncaughtException( Thread.currentThread(), e );
                 }
             }
         }, mainClass + ".main()" );
-        bootstrapThread.setContextClassLoader( getClassLoader() );
+        URLClassLoader classLoader = getClassLoader();
+        bootstrapThread.setContextClassLoader( classLoader );
         setSystemProperties();
 
         bootstrapThread.start();
@@ -326,6 +324,18 @@ public class ExecJavaMojo
             }
         }
 
+        if ( classLoader != null )
+        {
+            try
+            {
+                classLoader.close();
+            }
+            catch ( IOException e )
+            {
+                getLog().error(e.getMessage(), e);
+            }
+        }
+
         if ( originalSystemProperties != null )
         {
             System.setProperties( originalSystemProperties );
@@ -335,7 +345,7 @@ public class ExecJavaMojo
         {
             if ( threadGroup.uncaughtException != null )
             {
-                throw new MojoExecutionException( "An exception occured while executing the Java class. "
+                throw new MojoExecutionException( "An exception occurred while executing the Java class. "
                     + threadGroup.uncaughtException.getMessage(), threadGroup.uncaughtException );
             }
         }
@@ -343,6 +353,42 @@ public class ExecJavaMojo
         registerSourceRoots();
     }
 
+    /**
+     * To avoid the exec:java to consider common pool threads leaked, let's pre-create them.
+     */
+    private void preloadCommonPool()
+    {
+        try
+        {
+            // ensure common pool exists in the jvm
+            final ExecutorService es = ForkJoinPool.commonPool();
+            final int max = preloadCommonPool > 0
+                    ? preloadCommonPool :
+                    ForkJoinPool.getCommonPoolParallelism();
+            final CountDownLatch preLoad = new CountDownLatch( 1 );
+            for ( int i = 0;
+                  i < max;
+                  i++ )
+            {
+                es.submit(() -> {
+                    try
+                    {
+                        preLoad.await();
+                    }
+                    catch ( InterruptedException e )
+                    {
+                        Thread.currentThread().interrupt();
+                    }
+                });
+            }
+            preLoad.countDown();
+        }
+        catch (final Exception e)
+        {
+            getLog().debug(e.getMessage() + ", skipping commonpool earger init");
+        }
+    }
+
     /**
      * a ThreadGroup to isolate execution and collect exceptions.
      */
@@ -461,7 +507,7 @@ public class ExecJavaMojo
         }
         if ( !uncooperativeThreads.isEmpty() )
         {
-            getLog().warn( "NOTE: " + uncooperativeThreads.size() + " thread(s) did not finish despite being asked to "
+            getLog().warn( "NOTE: " + uncooperativeThreads.size() + " thread(s) did not finish despite being asked to"
                 + " via interruption. This is not a problem with exec:java, it is a problem with the running code."
                 + " Although not serious, it should be remedied." );
         }
@@ -513,37 +559,42 @@ public class ExecJavaMojo
      * @return the classloader
      * @throws MojoExecutionException if a problem happens
      */
-    private ClassLoader getClassLoader()
+    private URLClassLoader getClassLoader()
         throws MojoExecutionException
     {
-        List<URL> classpathURLs = new ArrayList<URL>();
+        List<Path> classpathURLs = new ArrayList<>();
         this.addRelevantPluginDependenciesToClasspath( classpathURLs );
         this.addRelevantProjectDependenciesToClasspath( classpathURLs );
         this.addAdditionalClasspathElements( classpathURLs );
-        return new URLClassLoader( classpathURLs.toArray( new URL[classpathURLs.size()] ) );
+        
+        try
+        {
+            return URLClassLoaderBuilder.builder()
+                    .setLogger( getLog() )
+                    .setPaths( classpathURLs )
+                    .setExclusions( classpathFilenameExclusions )
+                    .build();
+        }
+        catch ( NullPointerException | IOException e )
+        {
+            throw new MojoExecutionException( e.getMessage(), e );
+        }
+
     }
 
-    private void addAdditionalClasspathElements( List<URL> path )
+    private void addAdditionalClasspathElements( List<Path> path )
     {
         if ( additionalClasspathElements != null )
         {
             for ( String classPathElement : additionalClasspathElements )
             {
-                try
+                Path file = Paths.get( classPathElement );
+                if ( !file.isAbsolute() )
                 {
-                    File file = new File( classPathElement );
-                    if ( !file.isAbsolute() )
-                    {
-                        file = new File( project.getBasedir(), classPathElement );
-                    }
-                    URL url = file.toURI().toURL();
-                    getLog().debug( "Adding additional classpath element: " + url + " to classpath" );
-                    path.add( url );
-                }
-                catch ( MalformedURLException e )
-                {
-                    getLog().warn( "Skipping additional classpath element: " + classPathElement, e );
+                    file = project.getBasedir().toPath().resolve( file );
                 }
+                getLog().debug( "Adding additional classpath element: " + file + " to classpath" );
+                path.add( file );
             }
         }
     }
@@ -555,7 +606,7 @@ public class ExecJavaMojo
      * @param path classpath of {@link java.net.URL} objects
      * @throws MojoExecutionException if a problem happens
      */
-    private void addRelevantPluginDependenciesToClasspath( List<URL> path )
+    private void addRelevantPluginDependenciesToClasspath( List<Path> path )
         throws MojoExecutionException
     {
         if ( hasCommandlineArgs() )
@@ -563,20 +614,12 @@ public class ExecJavaMojo
             arguments = parseCommandlineArgs();
         }
 
-        try
+        for ( Artifact classPathElement : this.determineRelevantPluginDependencies() )
         {
-            for ( Artifact classPathElement : this.determineRelevantPluginDependencies() )
-            {
-                getLog().debug( "Adding plugin dependency artifact: " + classPathElement.getArtifactId()
-                    + " to classpath" );
-                path.add( classPathElement.getFile().toURI().toURL() );
-            }
+            getLog().debug( "Adding plugin dependency artifact: " + classPathElement.getArtifactId()
+                + " to classpath" );
+            path.add( classPathElement.getFile().toPath() );
         }
-        catch ( MalformedURLException e )
-        {
-            throw new MojoExecutionException( "Error during setting up classpath", e );
-        }
-
     }
 
     /**
@@ -585,38 +628,29 @@ public class ExecJavaMojo
      * @param path classpath of {@link java.net.URL} objects
      * @throws MojoExecutionException if a problem happens
      */
-    private void addRelevantProjectDependenciesToClasspath( List<URL> path )
+    private void addRelevantProjectDependenciesToClasspath( List<Path> path )
         throws MojoExecutionException
     {
         if ( this.includeProjectDependencies )
         {
-            try
+            getLog().debug( "Project Dependencies will be included." );
+
+            List<Artifact> artifacts = new ArrayList<>();
+            List<Path> theClasspathFiles = new ArrayList<>();
+
+            collectProjectArtifactsAndClasspath( artifacts, theClasspathFiles );
+
+            for ( Path classpathFile : theClasspathFiles )
             {
-                getLog().debug( "Project Dependencies will be included." );
-
-                List<Artifact> artifacts = new ArrayList<Artifact>();
-                List<File> theClasspathFiles = new ArrayList<File>();
-
-                collectProjectArtifactsAndClasspath( artifacts, theClasspathFiles );
-
-                for ( File classpathFile : theClasspathFiles )
-                {
-                    URL url = classpathFile.toURI().toURL();
-                    getLog().debug( "Adding to classpath : " + url );
-                    path.add( url );
-                }
-
-                for ( Artifact classPathElement : artifacts )
-                {
-                    getLog().debug( "Adding project dependency artifact: " + classPathElement.getArtifactId()
-                        + " to classpath" );
-                    path.add( classPathElement.getFile().toURI().toURL() );
-                }
-
+                getLog().debug( "Adding to classpath : " + classpathFile );
+                path.add( classpathFile );
             }
-            catch ( MalformedURLException e )
+
+            for ( Artifact classPathElement : artifacts )
             {
-                throw new MojoExecutionException( "Error during setting up classpath", e );
+                getLog().debug( "Adding project dependency artifact: " + classPathElement.getArtifactId()
+                    + " to classpath" );
+                path.add( classPathElement.getFile().toPath() );
             }
         }
         else
@@ -642,14 +676,13 @@ public class ExecJavaMojo
             if ( this.executableDependency == null )
             {
                 getLog().debug( "All Plugin Dependencies will be included." );
-                relevantDependencies = new HashSet<Artifact>( this.pluginDependencies );
+                relevantDependencies = new HashSet<Artifact>( this.getPluginDependencies() );
             }
             else
             {
                 getLog().debug( "Selected plugin Dependencies will be included." );
                 Artifact executableArtifact = this.findExecutableArtifact();
-                Artifact executablePomArtifact = this.getExecutablePomArtifact( executableArtifact );
-                relevantDependencies = this.resolveExecutableDependencies( executablePomArtifact );
+                relevantDependencies = this.resolveExecutableDependencies( executableArtifact );
             }
         }
         else
@@ -660,19 +693,6 @@ public class ExecJavaMojo
         return relevantDependencies;
     }
 
-    /**
-     * Get the artifact which refers to the POM of the executable artifact.
-     * 
-     * @param executableArtifact this artifact refers to the actual assembly.
-     * @return an artifact which refers to the POM of the executable artifact.
-     */
-    private Artifact getExecutablePomArtifact( Artifact executableArtifact )
-    {
-        return this.artifactFactory.createBuildArtifact( executableArtifact.getGroupId(),
-                                                         executableArtifact.getArtifactId(),
-                                                         executableArtifact.getVersion(), "pom" );
-    }
-
     /**
      * Resolve the executable dependencies for the specified project
      * 
@@ -683,39 +703,26 @@ public class ExecJavaMojo
     private Set<Artifact> resolveExecutableDependencies( Artifact executablePomArtifact )
         throws MojoExecutionException
     {
-
-        Set<Artifact> executableDependencies;
         try
         {
-            MavenProject executableProject =
-                this.projectBuilder.buildFromRepository( executablePomArtifact, this.remoteRepositories,
-                                                         this.localRepository );
+            ArtifactResolutionRequest request = new ArtifactResolutionRequest()
+                .setArtifact( executablePomArtifact )
+                .setLocalRepository( getSession().getLocalRepository() )
+                .setRemoteRepositories( getSession().getRequest().getRemoteRepositories() )
+                .setForceUpdate( getSession().getRequest().isUpdateSnapshots() )
+                .setOffline( getSession().isOffline() )
+                .setResolveTransitively( true );
 
-            // get all of the dependencies for the executable project
-            List<Dependency> dependencies = executableProject.getDependencies();
+            ArtifactResolutionResult result = repositorySystem.resolve( request );
+            resolutionErrorHandler.throwErrors( request, result );
 
-            // make Artifacts of all the dependencies
-            Set<Artifact> dependencyArtifacts =
-                MavenMetadataSource.createArtifacts( this.artifactFactory, dependencies, null, null, null );
-
-            // not forgetting the Artifact of the project itself
-            dependencyArtifacts.add( executableProject.getArtifact() );
-
-            // resolve all dependencies transitively to obtain a comprehensive list of assemblies
-            ArtifactResolutionResult result =
-                artifactResolver.resolveTransitively( dependencyArtifacts, executablePomArtifact,
-                                                      Collections.emptyMap(), this.localRepository,
-                                                      this.remoteRepositories, metadataSource, null,
-                                                      Collections.emptyList() );
-            executableDependencies = result.getArtifacts();
+            return result.getArtifacts();
         }
-        catch ( Exception ex )
+        catch ( ArtifactResolutionException ex )
         {
             throw new MojoExecutionException( "Encountered problems resolving dependencies of the executable "
                 + "in preparation for its execution.", ex );
         }
-
-        return executableDependencies;
     }
 
     /**
